package jalview.io.vcf;

import htsjdk.samtools.util.CloseableIterator;
import htsjdk.variant.variantcontext.Allele;
import htsjdk.variant.variantcontext.VariantContext;
import htsjdk.variant.vcf.VCFHeader;
import htsjdk.variant.vcf.VCFHeaderLine;
import htsjdk.variant.vcf.VCFHeaderLineCount;
import htsjdk.variant.vcf.VCFInfoHeaderLine;

import jalview.analysis.AlignmentUtils;
import jalview.analysis.Dna;
import jalview.api.AlignViewControllerGuiI;
import jalview.datamodel.AlignmentI;
import jalview.datamodel.DBRefEntry;
import jalview.datamodel.GeneLociI;
import jalview.datamodel.Mapping;
import jalview.datamodel.SequenceFeature;
import jalview.datamodel.SequenceI;
import jalview.ext.ensembl.EnsemblMap;
import jalview.ext.htsjdk.VCFReader;
import jalview.io.gff.Gff3Helper;
import jalview.io.gff.SequenceOntologyI;
import jalview.util.MapList;
import jalview.util.MappingUtils;
import jalview.util.MessageManager;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

/**
 * A class to read VCF data (using the htsjdk) and add variants as sequence
 * features on dna and any related protein product sequences
 * 
 * @author gmcarstairs
 */
public class VCFLoader
{
  /*
   * keys to fields of VEP CSQ consequence data
   * see https://www.ensembl.org/info/docs/tools/vep/vep_formats.html
   */
  private static final String ALLELE_NUM_KEY = "ALLELE_NUM"; // 0 (ref), 1...

  private static final String FEATURE_KEY = "Feature"; // Ensembl stable id

  /*
   * default VCF INFO key for VEP consequence data
   * NB this can be overridden running VEP with --vcf_info_field
   * - we don't handle this case (require CSQ identifier)
   */
  private static final String CSQ = "CSQ";

  /*
   * separator for fields in consequence data
   */
  private static final String PIPE = "|";

  private static final String PIPE_REGEX = "\\" + PIPE;

  /*
   * key for Allele Frequency output by VEP
   * see http://www.ensembl.org/info/docs/tools/vep/vep_formats.html
   */
  private static final String ALLELE_FREQUENCY_KEY = "AF";

  /*
   * delimiter that separates multiple consequence data blocks
   */
  private static final String COMMA = ",";

  /*
   * (temporary) flag that determines whether Jalview adds one feature
   * per VCF record, or one per allele (preferred)
   */
  private static final boolean FEATURE_PER_ALLELE = true;

  /*
   * the feature group assigned to a VCF variant in Jalview
   */
  private static final String FEATURE_GROUP_VCF = "VCF";

  /*
   * internal delimiter used to build keys for assemblyMappings
   * 
   */
  private static final String EXCL = "!";

  /*
   * the alignment we are associating VCF data with
   */
  private AlignmentI al;

  /*
   * mappings between VCF and sequence reference assembly regions, as 
   * key = "species!chromosome!fromAssembly!toAssembly
   * value = Map{fromRange, toRange}
   */
  private Map<String, Map<int[], int[]>> assemblyMappings;

  /*
   * holds details of the VCF header lines (metadata)
   */
  private VCFHeader header;

  /*
   * the position (0...) of the ALLELE_NUM field in each block of
   * CSQ (consequence) data (if declared in the VCF INFO header for CSQ)
   * see http://www.ensembl.org/info/docs/tools/vep/vep_formats.html
   */
  private int csqAlleleNumberFieldIndex = -1;

  private int csqFeatureFieldIndex = -1;

  /**
   * Constructor given an alignment context
   * 
   * @param alignment
   */
  public VCFLoader(AlignmentI alignment)
  {
    al = alignment;

    // map of species!chromosome!fromAssembly!toAssembly to {fromRange, toRange}
    assemblyMappings = new HashMap<String, Map<int[], int[]>>();
  }

  /**
   * Starts a new thread to query and load VCF variant data on to the alignment
   * <p>
   * This method is not thread safe - concurrent threads should use separate
   * instances of this class.
   * 
   * @param filePath
   * @param gui
   */
  public void loadVCF(final String filePath,
          final AlignViewControllerGuiI gui)
  {
    if (gui != null)
    {
      gui.setStatus(MessageManager.getString("label.searching_vcf"));
    }

    new Thread()
    {

      @Override
      public void run()
      {
        VCFLoader.this.doLoad(filePath, gui);
      }

    }.start();
  }

  /**
   * Loads VCF on to an alignment - provided it can be related to one or more
   * sequence's chromosomal coordinates
   * 
   * @param filePath
   * @param gui
   *          optional callback handler for messages
   */
  protected void doLoad(String filePath, AlignViewControllerGuiI gui)
  {
    VCFReader reader = null;
    try
    {
      // long start = System.currentTimeMillis();
      reader = new VCFReader(filePath);

      header = reader.getFileHeader();
      VCFHeaderLine ref = header
              .getOtherHeaderLine(VCFHeader.REFERENCE_KEY);

      /*
       * note offset of CSQ ALLELE_NUM field if it is declared
       */
      locateCsqFields();

      // check if reference is wrt assembly19 (GRCh37)
      // todo may need to allow user to specify reference assembly?
      boolean isRefGrch37 = (ref != null && ref.getValue().contains(
              "assembly19"));

      int varCount = 0;
      int seqCount = 0;

      /*
       * query for VCF overlapping each sequence in turn
       */
      for (SequenceI seq : al.getSequences())
      {
        int added = loadVCF(seq, reader, isRefGrch37);
        if (added > 0)
        {
          seqCount++;
          varCount += added;
          transferAddedFeatures(seq);
        }
      }
      if (gui != null)
      {
        // long elapsed = System.currentTimeMillis() - start;
        String msg = MessageManager.formatMessage("label.added_vcf",
                varCount, seqCount);
        gui.setStatus(msg);
        if (gui.getFeatureSettingsUI() != null)
        {
          gui.getFeatureSettingsUI().discoverAllFeatureData();
        }
      }
    } catch (Throwable e)
    {
      System.err.println("Error processing VCF: " + e.getMessage());
      e.printStackTrace();
      if (gui != null)
      {
        gui.setStatus("Error occurred - see console for details");
      }
    } finally
    {
      if (reader != null)
      {
        try
        {
          reader.close();
        } catch (IOException e)
        {
          // ignore
        }
      }
    }
  }

  /**
   * Records the position of fields for ALLELE_NUM and Feature defined in the
   * CSQ INFO header (if there is one). CSQ fields are declared in the CSQ INFO
   * Description e.g.
   * <p>
   * Description="Consequence ...from ... VEP. Format: Allele|Consequence|...
   */
  protected void locateCsqFields()
  {
    VCFInfoHeaderLine csqInfo = header.getInfoHeaderLine(CSQ);
    if (csqInfo == null)
    {
      return;
    }

    String desc = csqInfo.getDescription();
    if (desc != null)
    {
      String[] format = desc.split(PIPE_REGEX);
      int index = 0;
      for (String field : format)
      {
        if (ALLELE_NUM_KEY.equals(field))
        {
          csqAlleleNumberFieldIndex = index;
        }
        if (FEATURE_KEY.equals(field))
        {
          csqFeatureFieldIndex = index;
        }
        index++;
      }
    }
  }

  /**
   * Transfers VCF features to sequences to which this sequence has a mapping.
   * If the mapping is 1:3, computes peptide variants from nucleotide variants.
   * 
   * @param seq
   */
  protected void transferAddedFeatures(SequenceI seq)
  {
    DBRefEntry[] dbrefs = seq.getDBRefs();
    if (dbrefs == null)
    {
      return;
    }
    for (DBRefEntry dbref : dbrefs)
    {
      Mapping mapping = dbref.getMap();
      if (mapping == null || mapping.getTo() == null)
      {
        continue;
      }

      SequenceI mapTo = mapping.getTo();
      MapList map = mapping.getMap();
      if (map.getFromRatio() == 3)
      {
        /*
         * dna-to-peptide product mapping
         */
        AlignmentUtils.computeProteinFeatures(seq, mapTo, map);
      }
      else
      {
        /*
         * nucleotide-to-nucleotide mapping e.g. transcript to CDS
         */
        List<SequenceFeature> features = seq.getFeatures()
                .getPositionalFeatures(SequenceOntologyI.SEQUENCE_VARIANT);
        for (SequenceFeature sf : features)
        {
          if (FEATURE_GROUP_VCF.equals(sf.getFeatureGroup()))
          {
            transferFeature(sf, mapTo, map);
          }
        }
      }
    }
  }

  /**
   * Tries to add overlapping variants read from a VCF file to the given
   * sequence, and returns the number of variant features added. Note that this
   * requires the sequence to hold information as to its chromosomal positions
   * and reference, in order to be able to map the VCF variants to the sequence.
   * 
   * @param seq
   * @param reader
   * @param isVcfRefGrch37
   * @return
   */
  protected int loadVCF(SequenceI seq, VCFReader reader,
          boolean isVcfRefGrch37)
  {
    int count = 0;
    GeneLociI seqCoords = seq.getGeneLoci();
    if (seqCoords == null)
    {
      return 0;
    }

    List<int[]> seqChromosomalContigs = seqCoords.getMap().getToRanges();
    for (int[] range : seqChromosomalContigs)
    {
      count += addVcfVariants(seq, reader, range, isVcfRefGrch37);
    }

    return count;
  }

  /**
   * Queries the VCF reader for any variants that overlap the given chromosome
   * region of the sequence, and adds as variant features. Returns the number of
   * overlapping variants found.
   * 
   * @param seq
   * @param reader
   * @param range
   *          start-end range of a sequence region in its chromosomal
   *          coordinates
   * @param isVcfRefGrch37
   *          true if the VCF is with reference to GRCh37
   * @return
   */
  protected int addVcfVariants(SequenceI seq, VCFReader reader,
          int[] range, boolean isVcfRefGrch37)
  {
    GeneLociI seqCoords = seq.getGeneLoci();

    String chromosome = seqCoords.getChromosomeId();
    String seqRef = seqCoords.getAssemblyId();
    String species = seqCoords.getSpeciesId();

    /*
     * map chromosomal coordinates from GRCh38 (sequence) to
     * GRCh37 (VCF) if necessary
     */
    // TODO generalise for other assemblies and species
    int offset = 0;
    String fromRef = "GRCh38";
    if (fromRef.equalsIgnoreCase(seqRef) && isVcfRefGrch37)
    {
      String toRef = "GRCh37";
      int[] newRange = mapReferenceRange(range, chromosome, "human",
              fromRef, toRef);
      if (newRange == null)
      {
        System.err.println(String.format(
                "Failed to map %s:%s:%s:%d:%d to %s", species, chromosome,
                fromRef, range[0], range[1], toRef));
        return 0;
      }
      offset = newRange[0] - range[0];
      range = newRange;
    }

    boolean forwardStrand = range[0] <= range[1];

    /*
     * query the VCF for overlaps
     * (convert a reverse strand range to forwards)
     */
    int count = 0;
    MapList mapping = seqCoords.getMap();

    int fromLocus = Math.min(range[0], range[1]);
    int toLocus = Math.max(range[0], range[1]);
    CloseableIterator<VariantContext> variants = reader.query(chromosome,
            fromLocus, toLocus);
    while (variants.hasNext())
    {
      /*
       * get variant location in sequence chromosomal coordinates
       */
      VariantContext variant = variants.next();

      /*
       * we can only process SNP variants (which can be reported
       * as part of a MIXED variant record
       */
      if (!variant.isSNP() && !variant.isMixed())
      {
        continue;
      }

      int start = variant.getStart() - offset;
      int end = variant.getEnd() - offset;

      /*
       * convert chromosomal location to sequence coordinates
       * - null if a partially overlapping feature
       */
      int[] seqLocation = mapping.locateInFrom(start, end);
      if (seqLocation != null)
      {
        count += addVariantFeature(seq, variant, seqLocation[0],
                seqLocation[1], forwardStrand);
      }
    }

    variants.close();

    return count;
  }

  /**
   * Inspects the VCF variant record, and adds variant features to the sequence.
   * Only SNP variants are added, not INDELs. Returns the number of features
   * added.
   * <p>
   * If the sequence maps to the reverse strand of the chromosome, reference and
   * variant bases are recorded as their complements (C/G, A/T).
   * 
   * @param seq
   * @param variant
   * @param featureStart
   * @param featureEnd
   * @param forwardStrand
   */
  protected int addVariantFeature(SequenceI seq, VariantContext variant,
          int featureStart, int featureEnd, boolean forwardStrand)
  {
    byte[] reference = variant.getReference().getBases();
    if (reference.length != 1)
    {
      /*
       * sorry, we don't handle INDEL variants
       */
      return 0;
    }

    if (FEATURE_PER_ALLELE)
    {
      return addAlleleFeatures(seq, variant, featureStart, featureEnd,
              forwardStrand);
    }

    /*
     * for now we extract allele frequency as feature score; note
     * this attribute is String for a simple SNP, but List<String> if
     * multiple alleles at the locus; we extract for the simple case only
     */
    float score = getAlleleFrequency(variant, 0);

    StringBuilder sb = new StringBuilder();
    sb.append(forwardStrand ? (char) reference[0] : complement(reference));

    /*
     * inspect alleles and record SNP variants (as the variant
     * record could be MIXED and include INDEL and SNP alleles)
     * warning: getAlleles gives no guarantee as to the order 
     * in which they are returned
     */
    for (Allele allele : variant.getAlleles())
    {
      if (!allele.isReference())
      {
        byte[] alleleBase = allele.getBases();
        if (alleleBase.length == 1)
        {
          sb.append(COMMA).append(
                  forwardStrand ? (char) alleleBase[0]
                          : complement(alleleBase));
        }
      }
    }
    String alleles = sb.toString(); // e.g. G,A,C

    String type = SequenceOntologyI.SEQUENCE_VARIANT;

    SequenceFeature sf = new SequenceFeature(type, alleles, featureStart,
            featureEnd, score, FEATURE_GROUP_VCF);

    sf.setValue(Gff3Helper.ALLELES, alleles);

    Map<String, Object> atts = variant.getAttributes();
    for (Entry<String, Object> att : atts.entrySet())
    {
      sf.setValue(att.getKey(), att.getValue());
    }
    seq.addSequenceFeature(sf);

    return 1;
  }

  /**
   * A convenience method to get the AF value for the given alternate allele
   * index
   * 
   * @param variant
   * @param alleleIndex
   * @return
   */
  protected float getAlleleFrequency(VariantContext variant, int alleleIndex)
  {
    float score = 0f;
    String attributeValue = getAttributeValue(variant,
            ALLELE_FREQUENCY_KEY, alleleIndex);
    if (attributeValue != null)
    {
      try
      {
        score = Float.parseFloat(attributeValue);
      } catch (NumberFormatException e)
      {
        // leave as 0
      }
    }

    return score;
  }

  /**
   * A convenience method to get an attribute value for an alternate allele
   * 
   * @param variant
   * @param attributeName
   * @param alleleIndex
   * @return
   */
  protected String getAttributeValue(VariantContext variant,
          String attributeName, int alleleIndex)
  {
    Object att = variant.getAttribute(attributeName);

    if (att instanceof String)
    {
      return (String) att;
    }
    else if (att instanceof ArrayList)
    {
      return ((List<String>) att).get(alleleIndex);
    }

    return null;
  }

  /**
   * Adds one variant feature for each SNP allele in the VCF variant record, and
   * returns the number of features added.
   * 
   * @param seq
   * @param variant
   * @param featureStart
   * @param featureEnd
   * @param forwardStrand
   * @return
   */
  protected int addAlleleFeatures(SequenceI seq, VariantContext variant,
          int featureStart, int featureEnd, boolean forwardStrand)
  {
    int added = 0;

    /*
     * Javadoc says getAlternateAlleles() imposes no order on the list returned
     * so we proceed defensively to get them in strict order
     */
    int altAlleleCount = variant.getAlternateAlleles().size();
    for (int i = 0; i < altAlleleCount; i++)
    {
      added += addAlleleFeature(seq, variant, i, featureStart, featureEnd,
              forwardStrand);
    }
    return added;
  }

  /**
   * Inspects one allele and attempts to add a variant feature for it to the
   * sequence. Only SNP variants are added as features. We extract as much as
   * possible of the additional data associated with this allele to store in the
   * feature's key-value map. Answers the number of features added (0 or 1).
   * 
   * @param seq
   * @param variant
   * @param altAlleleIndex
   * @param featureStart
   * @param featureEnd
   * @param forwardStrand
   * @return
   */
  protected int addAlleleFeature(SequenceI seq, VariantContext variant,
          int altAlleleIndex, int featureStart, int featureEnd,
          boolean forwardStrand)
  {
    byte[] reference = variant.getReference().getBases();
    Allele alt = variant.getAlternateAllele(altAlleleIndex);
    byte[] allele = alt.getBases();
    if (allele.length != 1)
    {
      /*
       * not a SNP variant
       */
      return 0;
    }

    /*
     * build the ref,alt allele description e.g. "G,A"
     */
    StringBuilder sb = new StringBuilder();
    sb.append(forwardStrand ? (char) reference[0] : complement(reference));
    sb.append(COMMA);
    sb.append(forwardStrand ? (char) allele[0] : complement(allele));
    String alleles = sb.toString(); // e.g. G,A

    String type = SequenceOntologyI.SEQUENCE_VARIANT;
    float score = getAlleleFrequency(variant, altAlleleIndex);

    SequenceFeature sf = new SequenceFeature(type, alleles, featureStart,
            featureEnd, score, FEATURE_GROUP_VCF);

    sf.setValue(Gff3Helper.ALLELES, alleles);

    addAlleleProperties(variant, seq, sf, altAlleleIndex);

    seq.addSequenceFeature(sf);

    return 1;
  }

  /**
   * Add any allele-specific VCF key-value data to the sequence feature
   * 
   * @param variant
   * @param seq
   * @param sf
   * @param altAlelleIndex
   */
  protected void addAlleleProperties(VariantContext variant, SequenceI seq,
          SequenceFeature sf, final int altAlelleIndex)
  {
    Map<String, Object> atts = variant.getAttributes();

    for (Entry<String, Object> att : atts.entrySet())
    {
      String key = att.getKey();

      /*
       * extract Consequence data (if present) that we are able to
       * associated with the allele for this variant feature
       */
      if (CSQ.equals(key) && csqAlleleNumberFieldIndex > -1)
      {
        addConsequences(att.getValue(), seq, sf, altAlelleIndex + 1);
        return;
      }

      /*
       * we extract values for other data which are allele-specific; 
       * these may be per alternate allele (INFO[key].Number = 'A') 
       * or per allele including reference (INFO[key].Number = 'R') 
       */
      VCFInfoHeaderLine infoHeader = header.getInfoHeaderLine(key);
      if (infoHeader == null)
      {
        /*
         * can't be sure what data belongs to this allele, so
         * play safe and don't take any
         */
        continue;
      }

      VCFHeaderLineCount number = infoHeader.getCountType();
      int index = altAlelleIndex;
      if (number == VCFHeaderLineCount.R)
      {
        /*
         * one value per allele including reference, so bump index
         * e.g. the 3rd value is for the  2nd alternate allele
         */
        index++;
      }
      else if (number != VCFHeaderLineCount.A)
      {
        /*
         * don't save other values as not allele-related
         */
        continue;
      }

      /*
       * take the index'th value
       */
      String value = getAttributeValue(variant, key, index);
      if (value != null)
      {
        sf.setValue(key, value);
      }
    }
  }

  /**
   * Inspects CSQ data blocks (consequences) and adds attributes on the sequence
   * feature for the current allele (and transcript if applicable)
   * <p>
   * Allele matching: we require field ALLELE_NUM to match altAlleleIndex. If
   * the CSQ data does not include ALLELE_NUM values then no data is added to
   * the variant feature.
   * <p>
   * Transcript matching: if sequence name can be identified to at least one of
   * the consequences' Feature values, then select only consequences that match
   * the value (i.e. consequences for the current transcript sequence). If not,
   * take all consequences (this is the case when adding features to the gene
   * sequence).
   * 
   * @param value
   * @param seq
   * @param sf
   * @param altAlelleIndex
   *          (1=first alternative allele...)
   */
  protected void addConsequences(Object value, SequenceI seq,
          SequenceFeature sf, int altAlelleIndex)
  {
    if (!(value instanceof ArrayList<?>))
    {
      return;
    }

    List<String> consequences = (List<String>) value;

    /*
     * if CSQ data includes 'Feature', and any value matches the sequence name,
     * then restrict consequence data to the matching value (transcript)
     * i.e. just pick out consequences for the transcript the variant feature is on
     */
    String seqName = seq.getName()== null ? "" : seq.getName().toLowerCase();
    boolean matchFeature = false;
    String matchFeatureValue = null;
    if (csqFeatureFieldIndex > -1)
    {
      for (String consequence : consequences)
      {
        String[] csqFields = consequence.split(PIPE_REGEX);
        if (csqFields.length > csqFeatureFieldIndex)
        {
          String featureIdentifier = csqFields[csqFeatureFieldIndex];
          if (featureIdentifier.length() > 4
                  && seqName.indexOf(featureIdentifier.toLowerCase()) > -1)
          {
            matchFeature = true;
            matchFeatureValue = featureIdentifier;
          }
        }
      }
    }

    StringBuilder sb = new StringBuilder(128);
    boolean found = false;

    for (String consequence : consequences)
    {
      String[] csqFields = consequence.split(PIPE_REGEX);

      /*
       * check consequence is for the current transcript
       */
      if (matchFeature)
      {
        if (csqFields.length <= csqFeatureFieldIndex)
        {
          continue;
        }
        String featureIdentifier = csqFields[csqFeatureFieldIndex];
        if (!featureIdentifier.equals(matchFeatureValue))
        {
          continue; // consequence is for a different transcript
        }
      }

      if (csqFields.length > csqAlleleNumberFieldIndex)
      {
        String alleleNum = csqFields[csqAlleleNumberFieldIndex];
        if (String.valueOf(altAlelleIndex).equals(alleleNum))
        {
          if (found)
          {
            sb.append(COMMA);
          }
          found = true;
          sb.append(consequence);
        }
      }
    }

    if (found)
    {
      sf.setValue(CSQ, sb.toString());
    }
  }

  /**
   * A convenience method to complement a dna base and return the string value
   * of its complement
   * 
   * @param reference
   * @return
   */
  protected String complement(byte[] reference)
  {
    return String.valueOf(Dna.getComplement((char) reference[0]));
  }

  /**
   * Determines the location of the query range (chromosome positions) in a
   * different reference assembly.
   * <p>
   * If the range is just a subregion of one for which we already have a mapping
   * (for example, an exon sub-region of a gene), then the mapping is just
   * computed arithmetically.
   * <p>
   * Otherwise, calls the Ensembl REST service that maps from one assembly
   * reference's coordinates to another's
   * 
   * @param queryRange
   *          start-end chromosomal range in 'fromRef' coordinates
   * @param chromosome
   * @param species
   * @param fromRef
   *          assembly reference for the query coordinates
   * @param toRef
   *          assembly reference we wish to translate to
   * @return the start-end range in 'toRef' coordinates
   */
  protected int[] mapReferenceRange(int[] queryRange, String chromosome,
          String species, String fromRef, String toRef)
  {
    /*
     * first try shorcut of computing the mapping as a subregion of one
     * we already have (e.g. for an exon, if we have the gene mapping)
     */
    int[] mappedRange = findSubsumedRangeMapping(queryRange, chromosome,
            species, fromRef, toRef);
    if (mappedRange != null)
    {
      return mappedRange;
    }

    /*
     * call (e.g.) http://rest.ensembl.org/map/human/GRCh38/17:45051610..45109016:1/GRCh37
     */
    EnsemblMap mapper = new EnsemblMap();
    int[] mapping = mapper.getMapping(species, chromosome, fromRef, toRef,
            queryRange);

    if (mapping == null)
    {
      // mapping service failure
      return null;
    }

    /*
     * save mapping for possible future re-use
     */
    String key = makeRangesKey(chromosome, species, fromRef, toRef);
    if (!assemblyMappings.containsKey(key))
    {
      assemblyMappings.put(key, new HashMap<int[], int[]>());
    }

    assemblyMappings.get(key).put(queryRange, mapping);

    return mapping;
  }

  /**
   * If we already have a 1:1 contiguous mapping which subsumes the given query
   * range, this method just calculates and returns the subset of that mapping,
   * else it returns null. In practical terms, if a gene has a contiguous
   * mapping between (for example) GRCh37 and GRCh38, then we assume that its
   * subsidiary exons occupy unchanged relative positions, and just compute
   * these as offsets, rather than do another lookup of the mapping.
   * <p>
   * If in future these assumptions prove invalid (e.g. for bacterial dna?!),
   * simply remove this method or let it always return null.
   * <p>
   * Warning: many rapid calls to the /map service map result in a 429 overload
   * error response
   * 
   * @param queryRange
   * @param chromosome
   * @param species
   * @param fromRef
   * @param toRef
   * @return
   */
  protected int[] findSubsumedRangeMapping(int[] queryRange, String chromosome,
          String species, String fromRef, String toRef)
  {
    String key = makeRangesKey(chromosome, species, fromRef, toRef);
    if (assemblyMappings.containsKey(key))
    {
      Map<int[], int[]> mappedRanges = assemblyMappings.get(key);
      for (Entry<int[], int[]> mappedRange : mappedRanges.entrySet())
      {
        int[] fromRange = mappedRange.getKey();
        int[] toRange = mappedRange.getValue();
        if (fromRange[1] - fromRange[0] == toRange[1] - toRange[0])
        {
          /*
           * mapping is 1:1 in length, so we trust it to have no discontinuities
           */
          if (MappingUtils.rangeContains(fromRange, queryRange))
          {
            /*
             * fromRange subsumes our query range
             */
            int offset = queryRange[0] - fromRange[0];
            int mappedRangeFrom = toRange[0] + offset;
            int mappedRangeTo = mappedRangeFrom + (queryRange[1] - queryRange[0]);
            return new int[] { mappedRangeFrom, mappedRangeTo };
          }
        }
      }
    }
    return null;
  }

  /**
   * Transfers the sequence feature to the target sequence, locating its start
   * and end range based on the mapping. Features which do not overlap the
   * target sequence are ignored.
   * 
   * @param sf
   * @param targetSequence
   * @param mapping
   *          mapping from the feature's coordinates to the target sequence
   */
  protected void transferFeature(SequenceFeature sf,
          SequenceI targetSequence, MapList mapping)
  {
    int[] mappedRange = mapping.locateInTo(sf.getBegin(), sf.getEnd());
  
    if (mappedRange != null)
    {
      String group = sf.getFeatureGroup();
      int newBegin = Math.min(mappedRange[0], mappedRange[1]);
      int newEnd = Math.max(mappedRange[0], mappedRange[1]);
      SequenceFeature copy = new SequenceFeature(sf, newBegin, newEnd,
              group, sf.getScore());
      targetSequence.addSequenceFeature(copy);
    }
  }

  /**
   * Formats a ranges map lookup key
   * 
   * @param chromosome
   * @param species
   * @param fromRef
   * @param toRef
   * @return
   */
  protected static String makeRangesKey(String chromosome, String species,
          String fromRef, String toRef)
  {
    return species + EXCL + chromosome + EXCL + fromRef + EXCL
            + toRef;
  }
}
